from .logger import get_logger
from .connect import get_connection
from cassandra.query import SimpleStatement
import pandas as pd
import os

_dict_cql_to_python_type = {
    'NULL': None,
    'boolean': pd.BooleanDtype(),
    'float': pd.Float64Dtype(),
    'double': pd.Float64Dtype(),
    'int': pd.Int64Dtype(),
    'smallint': pd.Int16Dtype(),
    'tinyint': pd.Int64Dtype(),
    'counter': pd.Int64Dtype(),
    'varint': pd.Int64Dtype(),
    'bigint': pd.Int64Dtype(),
    'decimal': pd.Float64Dtype(),
    'ascii': str,
    'varchar': str,
    'text': str,
    'blob': str,
    'date': str,
    'timestamp': str,
    'time': str,
    'list': list,
    'set': set,
    'map': dict,
    'timeuuid,': None,
    'uuid': None
}

logger = get_logger(os.path.basename(__file__)+"_get_schema_cols")


def get_schema_cols(hosts: list, port: int, keyspace: str, table_name: str, username: str, password: str):
    cluster, session = get_connection(
        hosts, port, keyspace, username, password)
    statment = SimpleStatement(
        "SELECT column_name,type FROM system_schema.columns WHERE keyspace_name=%s AND table_name=%s")
    logger.info(f"Start fetching columns from the {table_name}")
    data = session.execute(statment, (keyspace, table_name))
    list_table_cols = []
    dict_csv_cols = {}
    for row in data:
        dict_csv_cols[row.column_name] = _dict_cql_to_python_type[row.type]
        list_table_cols.append(row.column_name)
    session.shutdown()
    cluster.shutdown()
    logger.debug("Successfully shutdown session.")
    return list_table_cols, dict_csv_cols


__all__ = ["get_schema_cols", "cql_to_python_type"]



import logging
import sys
logging.getLogger("cassandra").setLevel('ERROR')


def get_logger(name: str):
    # Create a logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Create a file handler and set the logging level
    file_handler = logging.FileHandler(f'{name}.log')
    file_handler.setLevel(logging.DEBUG)

    # Create a console handler and set the logging level
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)

    # Create a formatter and set it for the handlers
    formatter = logging.Formatter(
        '%(asctime)-15s | %(name)-26s | %(levelname)-8s | %(message)s')
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)
    logger.propagate = False
    return logger


__all__ = ["get_logger"]



from .logger import get_logger
from cassandra.cluster import Cluster
from cassandra.auth import PlainTextAuthProvider
import os

logger = get_logger(os.path.basename(__file__)+"_get_connection")


def get_connection(hosts: list, port: int, keyspace: str, username: str = 'cassandra', password: str = 'cassandra'):
    try:
        auth = PlainTextAuthProvider(username=username, password=password)
        cluster = Cluster(contact_points=hosts, port=port, auth_provider=auth)
        logger.info(f"Connection establised...{str(hosts)}:{port}")
        session = cluster.connect(keyspace)
        logger.info(f"Keyspace: {keyspace} is conncted.")
        return cluster, session
    except Exception as e:
        logger.error("\n{}\nERROR".format('-'*30), exc_info=e)
        exit(1)


__all__ = ["get_connection"]



from .logger import get_logger
from .connect import get_connection
from .schema import get_schema_cols
import pandas as pd
import os
import sys
import numpy as np
import threading
import re
import time
logger = get_logger(os.path.basename(__file__)+"_insert")


def cdc_insert(hosts: list, port: int = 9042, keyspace_name: str = '', username: str = 'cassandra', password: str = 'cassandra', file_name: str = 'data.csv', chunks: int = 10000, workers: int = 4) -> None:
    try:
        start_time = time.time()
        data_frames = pd.read_csv(file_name, chunksize=chunks)
        # find schema & table name from metadata column
        schema = [i.split("=")[1] for i in re.findall(
            "(?:schema=[A-z0-9\-_]+)|table=[A-z0-9\-_]+", next(data_frames).loc[0, "metadata"])]
        table_name = schema[0]

        cluster, session = get_connection(
            hosts, port, keyspace_name, username, password)
        list_columns, csv_dtypes = get_schema_cols(
            hosts, port, keyspace_name, table_name, username, password)

        if len(list_columns) <= 0:
            df = next(data_frames)
            df = df.drop('metadata', axis=1)  # drop metadata column
            list_columns = map(str.lower, df.columns.to_list())
            logger.warning(f"Unable to fetch columns from {table_name}")
            logger.info("Columns fetched from the CSV file")
        else:
            data_frames.close()
            data_frames = pd.read_csv(
                file_name, chunksize=chunks, dtype=csv_dtypes)
            logger.info(f"Columns successfully fetched from the {table_name}")

        insert_statement = "INSERT INTO "+table_name + \
            " (" + ",".join(list_columns) + ") VALUES (" + \
            ",".join(['%('+k+')s' for k in list_columns])+")"

        def execute_insert(df: pd.DataFrame, session) -> None:
            # convert lower names
            df.columns = map(str.lower, df.columns)
            # handle null for numeric type default 0 add
            df[df.select_dtypes(include=np.number).columns] = df.select_dtypes(
                include=np.number).fillna(0).astype(int)
            # fill boolean
            df[df.select_dtypes(include=bool).columns] = df.select_dtypes(
                include=bool).fillna(False)
            # handle null for object and string types
            df[df.select_dtypes(exclude=np.number).columns] = df.select_dtypes(
                exclude=np.number).fillna('')

            # empty list
            futures = []
            # excute async & store future object
            for row in df.to_dict(orient='records'):  # iterate each records
                futures.append(session.execute_async(insert_statement, row))

            # wait for complete async
            for future in futures:
                future.result()  # wait until insert queries

        threads = list(range(workers))
        logger.debug(f"{workers} workers are ready!")
        count = 0
        while True:
            try:
                # start thread
                for i in range(workers):
                    # fetch next dataframe
                    df = next(data_frames)
                    count += len(df)
                    try:
                        threads[i] = threading.Thread(
                            target=execute_insert, args=(df, session,))
                        threads[i].start()
                    except Exception as e:
                        logger.error("\n{}\nERROR".format('-'*30), exc_info=e)
                # wait until complete task
                for i in range(workers):
                    threads[i].join()
            # on execption break
            except Exception as e:
                break

        duration_minutes = (time.time() - start_time) / 60
        logger.debug(
            f"DONE! {count} Records Transferred In {duration_minutes:.2f} Minutes.")
    except Exception as e:
        logger.error("\n{}\nERROR".format('-'*30), exc_info=e)
        exit(1)


__all__ = ["cdc_insert"]



from .logger import get_logger
from .connect import get_connection
from .schema import get_schema_cols
import pandas as pd
import os
import sys
import numpy as np
import threading
import time
logger = get_logger(os.path.basename(__file__)+"_insert")


def bulk_insert(hosts: list, port: int = 9042, table_name: str = '', username: str = 'cassandra', password: str = 'cassandra', file_name: str = 'data.csv', chunks: int = 10000, workers: int = 4) -> None:
    try:
        start_time = time.time()
        data_frames = pd.read_csv(file_name, chunksize=chunks)
        if len(sys.argv) <= 1:
            if table_name == '':
                logger.error("Table name is not inputed!")
                exit(1)
        else:
            table_name = sys.argv[1]

        if '.' not in table_name:
            logger.error("Table name format must be keyspace_name.table_name!")
            exit(1)

        keyspace_name, table_name = table_name.split('.')
        cluster, session = get_connection(
            hosts, port, keyspace_name, username, password)
        list_columns, csv_dtypes = get_schema_cols(
            hosts, port, keyspace_name, table_name, username, password)

        if len(list_columns) <= 0:
            df = next(data_frames)
            list_columns = map(str.lower, df.columns.to_list())
            logger.warning(f"Unable to fetch columns from {table_name}")
            logger.info("Columns fetched from the CSV file")
        else:
            data_frames.close()
            data_frames = pd.read_csv(
                file_name, chunksize=chunks, dtype=csv_dtypes)
            logger.info(f"Columns successfully fetched from the {table_name}")

        insert_statement = "INSERT INTO "+table_name + \
            " (" + ",".join(list_columns) + ") VALUES (" + \
            ",".join(['%('+k+')s' for k in list_columns])+")"

        def execute_insert(df: pd.DataFrame, session) -> None:
            # convert lower names
            df.columns = map(str.lower, df.columns)
            # handle null for numeric type default 0 add
            df[df.select_dtypes(include=np.number).columns] = df.select_dtypes(
                include=np.number).fillna(0).astype(int)
            # fill boolean
            df[df.select_dtypes(include=bool).columns] = df.select_dtypes(
                include=bool).fillna(False)
            # handle null for object and string types
            df[df.select_dtypes(exclude=np.number).columns] = df.select_dtypes(
                exclude=np.number).fillna('')

            # empty list
            futures = []
            # excute async & store future object
            for row in df.to_dict(orient='records'):  # iterate each records
                # print(row)
                futures.append(session.execute_async(insert_statement, row))

            # wait for complete async
            for future in futures:
                future.result()  # wait until insert queries

        threads = list(range(workers))
        logger.debug(f"{workers} workers are ready!")
        count = 0
        while True:
            try:
                # start thread
                for i in range(workers):
                    # fetch next dataframe
                    df = next(data_frames)
                    count += len(df)
                    threads[i] = threading.Thread(
                        target=execute_insert, args=(df, session,))
                    threads[i].start()

                # wait until complete task
                for i in range(workers):
                    threads[i].join()
            # on execption break
            except Exception as e:
                break

        duration_minutes = (time.time() - start_time) / 60
        logger.debug(
            f"DONE! {count} Records Transferred In {duration_minutes:.2f} Minutes.")
    except Exception as e:
        logger.error("\n{}\nERROR".format('-'*30), exc_info=e)
        exit(1)


__all__ = ["bulk_insert"]
